from pronunciation_dictionary.deserialization import DeserializationOptions, deserialize
from pronunciation_dictionary.io import load_dict, load_dict_from_url, save_dict
from pronunciation_dictionary.mp_options import MultiprocessingOptions
from pronunciation_dictionary.phoneme_set_extraction import get_phoneme_set
from pronunciation_dictionary.pronunciation_selection import (get_first_pronunciation,
                                                              get_last_pronunciation,
                                                              get_longest_pronunciation,
                                                              get_pronunciation_with_highest_weight,
                                                              get_pronunciation_with_lowest_weight,
                                                              get_random_pronunciation,
                                                              get_shortest_pronunciation,
                                                              get_weighted_pronunciation)
from pronunciation_dictionary.serialization import SerializationOptions, serialize
from pronunciation_dictionary.types import (Pronunciation, PronunciationDict, Pronunciations,
                                            Symbol, Weight, Word)
from pronunciation_dictionary.validation import (validate_dictionary, validate_pronunciation,
                                                 validate_pronunciations, validate_seed,
                                                 validate_weight, validate_word)



from typing import Set

from pronunciation_dictionary.types import PronunciationDict, Symbol
from pronunciation_dictionary.validation import validate_dictionary


def get_phoneme_set(dictionary: PronunciationDict) -> Set[Symbol]:
  try:
    validate_dictionary(dictionary)
  except ValueError as error:
    raise ValueError("dictionary", error.args[1]) from error

  unique_symbols = {
    symbol
    for pronunciations_weight_pair in dictionary.values()
    for pronunciation in pronunciations_weight_pair.keys()
    for symbol in pronunciation
  }

  return unique_symbols



import re
from collections import OrderedDict
from dataclasses import dataclass
from functools import partial
from logging import getLogger
from multiprocessing.pool import Pool
from typing import List, Optional, Tuple

from pronunciation_dictionary.mp_options import MultiprocessingOptions
from pronunciation_dictionary.types import Pronunciation, PronunciationDict, Weight, Word
from pronunciation_dictionary.validation import validate_mp_options, validate_type

WORD_PRON_PATTERN = re.compile(r"(\S+)\s+(.+)")
WORD_WEIGHT_PRON_PATTERN = re.compile(r"(\S+)\s+([0-9\.]+)\s+(.+)")

# e.g. `ABBE(1)`
WORD_ALT_PATTERN = re.compile(r"(\S+)(\([0-9]+\))")
PRON_COMMENT_PATTERN = re.compile(r"(.*\S+)\s+(#.*)")
PRON_SYMB_SEP_PATTERN = re.compile(r"\s+")
DEFAULT_WEIGHT: Weight = 1.0


@dataclass()
class DeserializationOptions():
  consider_comments: bool
  consider_word_nrs: bool
  consider_pronunciation_comments: bool
  consider_weights: bool


def validate_deserialization_options(options: DeserializationOptions) -> Optional[str]:
  if msg := validate_type(options.consider_comments, bool):
    return f"Property 'consider_comments': {msg}"
  if msg := validate_type(options.consider_word_nrs, bool):
    return f"Property 'consider_word_nrs': {msg}"
  if msg := validate_type(options.consider_pronunciation_comments, bool):
    return f"Property 'consider_pronunciation_comments': {msg}"
  if msg := validate_type(options.consider_weights, bool):
    return f"Property 'consider_weights': {msg}"
  return None


def deserialize(lines: List[str], options: DeserializationOptions, mp_options: MultiprocessingOptions) -> PronunciationDict:
  if msg := validate_type(lines, list):
    raise ValueError(f"Property 'lines': {msg}")
  if msg := validate_deserialization_options(options):
    raise ValueError(f"Parameter 'options': {msg}")
  if msg := validate_mp_options(mp_options):
    raise ValueError(f"Parameter 'mp_options': {msg}")

  if len(lines) == 0:
    return OrderedDict()

  logger = getLogger(__name__)

  process_method = partial(
    process_get_pronunciation,
    options=options,
  )

  with Pool(
    processes=mp_options.n_jobs,
    initializer=__init_pool_prepare_cache_mp,
    initargs=(lines,),
    maxtasksperchild=mp_options.maxtasksperchild,
  ) as pool:
    entries = range(len(lines))
    iterator = pool.imap(process_method, entries, mp_options.chunksize)
    # iterator = tqdm(iterator, total=len(entries), unit="lines")
    result = dict(iterator)

  pronunciation_dict: PronunciationDict = OrderedDict()
  for line_i in range(len(lines)):
    line_nr = line_i + 1
    assert line_i in result
    values, messages = result[line_i]
    for message in messages:
      logger.info(f"Line {line_nr}: {message}")
    if values is None:
      continue
    word, weight, pronunciation = values
    had_weight = weight is not None

    if weight is None:
      weight = DEFAULT_WEIGHT

    if had_weight and weight == 0:
      logger.info(
        f"Line {line_nr}: Ignored line because to word \"{word}\" the pronunciation \"{' '.join(pronunciation)}\" had zero weight.")
    if word in pronunciation_dict:
      if pronunciation in pronunciation_dict[word]:
        if had_weight:
          existing_weight = pronunciation_dict[word]
          if weight != existing_weight:
            logger.warning(
              f"Line {line_nr}: Ignored line because to word \"{word}\" the pronunciation \"{' '.join(pronunciation)}\" was already assigned previously but with another weight ({existing_weight} vs. {weight})!.")
          else:
            logger.info(
              f"Line {line_nr}: Ignored line because to word \"{word}\" the pronunciation \"{' '.join(pronunciation)}\" was already assigned previously (with same weight of {weight}).")
        else:
          logger.info(
            f"Line {line_nr}: Ignored line because to word \"{word}\" the pronunciation \"{' '.join(pronunciation)}\" was already assigned previously.")
        continue
      pronunciation_dict[word][pronunciation] = weight
    else:
      pronunciation_dict[word] = OrderedDict((
        (pronunciation, weight),
      ))

  return pronunciation_dict


process_lines: List[str] = None


def __init_pool_prepare_cache_mp(lines: List[str]) -> None:
  global process_lines
  process_lines = lines


def process_get_pronunciation(line_i: int, options: DeserializationOptions) -> Tuple[int, Optional[Tuple[Tuple[Word, Optional[Weight], Pronunciation], List[str]]]]:
  global process_lines
  assert 0 <= line_i < len(process_lines)
  line = process_lines[line_i]
  return line_i, parse_line(line, options)


def parse_line(line: str, options: DeserializationOptions) -> Optional[Tuple[Tuple[Word, Optional[Weight], Pronunciation], List[str]]]:
  line = line.strip()
  line_is_empty = line == ""
  if line_is_empty:
    return None, ["Ignored empty line."]

  if options.consider_comments:
    is_comment = line.startswith(";;;")
    if is_comment:
      return None, [f"Ignored comment -> \"{line}\""]

  if options.consider_weights:
    word_weight_pronun_match = re.fullmatch(WORD_WEIGHT_PRON_PATTERN, line)
    is_invalid_line = word_weight_pronun_match is None
    if is_invalid_line:
      return None, [f"Ignored invalid line -> \"{line}\""]
    word = word_weight_pronun_match.group(1)
    weight = word_weight_pronun_match.group(2)
    pronunciation = word_weight_pronun_match.group(3)
    try:
      weight = float(weight)
    except ValueError as error:
      return None, [f"Weight couldn't be parsed -> \"{weight}\""]
  else:
    word_pronun_match = re.fullmatch(WORD_PRON_PATTERN, line)
    is_invalid_line = word_pronun_match is None
    if is_invalid_line:
      return None, [f"Ignored invalid line -> \"{line}\""]
    word = word_pronun_match.group(1)
    weight = None
    pronunciation = word_pronun_match.group(2)

  msgs = []

  if options.consider_word_nrs:
    word_nr_match = re.fullmatch(WORD_ALT_PATTERN, word)
    has_word_nr = word_nr_match is not None
    if has_word_nr:
      word = word_nr_match.group(1)
      msgs.append(f"Got alternate pronunciation \"{word_nr_match.group(2)}\" for word \"{word}\"")

  if options.consider_pronunciation_comments:
    comment_match = re.fullmatch(PRON_COMMENT_PATTERN, pronunciation)
    has_comment = comment_match is not None
    if has_comment:
      pronunciation = comment_match.group(1)
      msgs.append(
        f"Got comment for word \"{word}\" and pronunciation \"{pronunciation}\" -> \"{comment_match.group(2)}\"")

  symbols = re.split(PRON_SYMB_SEP_PATTERN, pronunciation)
  pronunciation = tuple(symbols)
  return (word, weight, pronunciation), msgs



from dataclasses import dataclass
from typing import Generator, Literal, Optional

from pronunciation_dictionary.types import Pronunciation, PronunciationDict, Pronunciations
from pronunciation_dictionary.validation import validate_dictionary

_PHONEME_SEP = " "


@dataclass()
class SerializationOptions():
  parts_sep: Literal["TAB", "SPACE", "DOUBLE-SPACE"]
  include_counter: bool
  include_weights: bool


_part_separators = {
  "TAB": "\t",
  "SPACE": " ",
  "DOUBLE-SPACE": "  ",
}


def _validate_serialization_options(options: SerializationOptions) -> Optional[str]:
  if options.parts_sep not in ["TAB", "SPACE", "DOUBLE-SPACE"]:
    return "Property 'parts_sep' is invalid!"
  if not isinstance(options.include_counter, bool):
    return "Property 'include_counter' is invalid!"
  if not isinstance(options.include_weights, bool):
    return "Property 'include_weights' is invalid!"
  return None


def serialize(dictionary: PronunciationDict, options: SerializationOptions) -> Generator[str, None, None]:
  try:
    validate_dictionary(dictionary)
  except ValueError as error:
    raise ValueError("dictionary", error.args[1]) from error

  if msg := _validate_serialization_options(options):
    raise ValueError("options", msg)

  part_separator = _part_separators[options.parts_sep]
  for word, pronunciations in dictionary.items():
    yield from _get_lines_for_pronunciation(word, pronunciations, part_separator, options.include_weights, options.include_counter)


def _get_lines_for_pronunciation(word: str, pronunciations: Pronunciations, part_separator: str, include_weights: bool, include_counter: bool) -> Generator[str, None, None]:
  for counter, (pronunciation, weight) in enumerate(pronunciations.items(), start=1):
    yield _get_line_for_pronunciation(word, pronunciation, counter, include_counter, part_separator, weight, include_weights)


def _get_line_for_pronunciation(word: str, pronunciation: Pronunciation, counter: int, include_counter: bool, part_separator: str, weight: float, include_weights: bool):
  word_part = word
  if include_counter:
    word_part = _get_word_serialized_with_counter(word, counter)
  weights_part = ""
  if include_weights:
    weights_part = f"{_get_weight_serialized(weight)}{part_separator}"
  pron_part = _get_pronunciation_serialized(pronunciation)
  line = f"{word_part}{part_separator}{weights_part}{pron_part}"
  return line


def _get_weight_serialized(weight: float) -> str:
  result = str(weight)
  return result


def _get_word_serialized_with_counter(word: str, counter: int) -> str:
  assert counter > 0
  counter_str = f"({counter})" if counter > 1 else ""
  word_part = f"{word}{counter_str}"
  return word_part


def _get_pronunciation_serialized(pronunciation: Pronunciation) -> str:
  pron_part = _PHONEME_SEP.join(pronunciation)
  return pron_part



from collections import OrderedDict
from typing import Any, Optional, Union

from pronunciation_dictionary.mp_options import MultiprocessingOptions
from pronunciation_dictionary.types import Pronunciation, PronunciationDict, Pronunciations


class ValidationError():
  # pylint: disable=no-self-use
  @property
  def default_message(self) -> str:
    return ""


class InternalError(ValidationError):
  @property
  def default_message(self) -> str:
    return "Internal error!"


def _contain_whitespace(chars: str) -> bool:
  for char in chars:
    if char in (" ", "\t"):
      return True
  return False


def validate_dictionary(dictionary: PronunciationDict) -> None:
  if not (isinstance(dictionary, OrderedDict)):
    raise ValueError("dictionary", "Type needs to be 'OrderedDict'!")
  for word, pronunciations in dictionary.items():
    try:
      validate_word(word)
    except ValueError as error:
      raise ValueError("dictionary", error.args[1]) from error
    if not isinstance(pronunciations, OrderedDict):
      raise ValueError("dictionary", "Pronunciations need to be of type 'OrderedDict'!")
    for pronunciation, weight in pronunciations.items():
      try:
        validate_pronunciation(pronunciation)
      except ValueError as error:
        raise ValueError("dictionary", error.args[1]) from error
      try:
        validate_weight(weight)
      except ValueError as error:
        raise ValueError("dictionary", error.args[1]) from error


def validate_word(word: str) -> None:
  if not isinstance(word, str):
    raise ValueError("word", "Word need to be of type 'str'!")
  if len(word) == 0:
    raise ValueError("word", "Empty words are not allowed!")
  if _contain_whitespace(word):
    raise ValueError("word", "Word contains whitespace which is not allowed!")


def validate_pronunciation(pronunciation: Pronunciation) -> None:
  if not isinstance(pronunciation, tuple):
    raise ValueError("pronunciation", "Pronunciation need to be of type 'tuple'!")

  if len(pronunciation) == 0:
    raise ValueError("pronunciation", "Pronunciation is empty!")

  all_phonemes_are_str = all(isinstance(phoneme, str) for phoneme in pronunciation)
  if not all_phonemes_are_str:
    raise ValueError("pronunciation", "Phonemes need to be of type 'str'!")

  whitespace_in_any_phoneme = any(
    _contain_whitespace(phoneme)
    for phoneme in pronunciation
  )
  if whitespace_in_any_phoneme:
    raise ValueError("pronunciation", "Pronunciation contains whitespace which is not allowed!")


def validate_weight(weight: Union[float, int]) -> None:
  if not isinstance(weight, (float, int)):
    raise ValueError("weight", "Weight needs to be of type 'float' or 'int'!")


def validate_mp_options(mp_options: MultiprocessingOptions) -> Optional[str]:
  if not (isinstance(mp_options.chunksize, int) and mp_options.chunksize > 0):
    return "Property 'chunksize' is invalid!"
  if not (mp_options.maxtasksperchild is None or (isinstance(mp_options.maxtasksperchild, int) and mp_options.maxtasksperchild > 0)):
    return "Property 'maxtasksperchild' is invalid!"
  if not (isinstance(mp_options.n_jobs, int) and mp_options.n_jobs > 0):
    return "Property 'n_jobs' is invalid!"
  return None


def validate_seed(seed: int) -> Optional[str]:
  if not 0 < seed and not isinstance(seed, int):
    return "Invalid value!"
  return None


def validate_type(obj: Any, t: type) -> Optional[str]:
  if not isinstance(obj, t):
    return f"Value needs of type '{t.__name__}'!"
  return None


def validate_pronunciations(pronunciations: Pronunciations) -> Optional[str]:
  if msg := validate_type(pronunciations, OrderedDict):
    return msg
  if not len(pronunciations) > 0:
    return "At least one pronunciation is required!"
  return None
